#-----------------------------------------------------------------------------------------------------
# mdlog/mu.py
#-----------------------------------------------------------------------------------------------------
import os.path, sys
from os import makedirs
#-----------------------------------------------------------------------------------------------------


class LOG:
    DEFAULT_FILE_EXT = '.html'
    """ Mark-Up Logging 
            ~> Log your outputs and results directly to mark-down format.
            ~> Generate on-the-fly reports, ready for presentation.
    """
    
    def __init__(self, log_dir, log_file):
        """
            log_dir         :   directory to create new log file at
            log_file        :   name of new log file, *AUTO ADDS '.md' EXTENSION*
            
            Note: By default, the escmode is False, does not escape any md-special chars
        """
        self.log_dir = log_dir              
        self.log_file = log_file + (  '' if log_file.lower().endswith(self.DEFAULT_FILE_EXT) else self.DEFAULT_FILE_EXT) 
        self.log_path = os.path.join(self.log_dir, self.log_file)
        #self.uri_title_start, self.uri_title_end = (('"', '"') if uri_title_quote else ('(', ')'))
        self.iomode='' #<--- no iomode means closeed
        self.escmode(False) #<---- manually set the escape mode after init if its required

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    
    """ Logger File Handles """
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
    def open(self, mode='w'):
        """ open log file handle """
        assert (mode=='a' or mode=='w')
        self.iomode = mode
        makedirs(self.log_dir, exist_ok=True)
        self.f = open(self.log_path, mode)

    def close(self):
        """ close log file handle """
        self.f.close()
        self.iomode=''
        del self.f

    def loc(self, file):
        """ returns relative loaction of a file wrt log file 
            - this is useful to linking local files in markdown
            - 'uri' function will auto-convert to relative path if its 'loc' arg is True
        """
        return os.path.relpath( file , self.log_dir )

    def info(self, p=print):
        """ short info about file handle"""
        p('[Logging ~ Mode:[{}] @File:[{}] @Path:[{}]'.format(self.iomode, self.log_file, self.log_path))

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    
    """  IO and Escaping """
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ 
    def escmode(self, esc):
        """ enable or disable escaping, sets the 'set of escape charaters' in self.esc 
            - if 'esc' evaluates to False, then escaping is disabled 
            - otherwise, it should be a tuple, all chars in the 'esc' tuple are escaped """
        self.esc = esc
        self.do_esc = True if self.esc else False
        self.write = self.write_do_esc if self.do_esc else self.write_no_esc

    def escape(self, msg):
        """ escapes all instances of chars in self.esc tuple """
        m = str(msg)
        for k,v in self.esc.items():
            m = m.replace(k, v)
        return m

    def escaper(self, escT):
        """ returns a context-manager for temporary escaping special chars - see the ESCAPER class """
        return ESCAPER((self, escT))

    def write_no_esc(self, *msg):
        """ write msg without escaping """
        for m in msg:
            self.f.write(str(m))

    def write_do_esc(self, *msg):  
        """ write msg with escaping - escapes all chars that are currently in self.esc tuple """
        emsg = map(self.escape, msg)
        for m in emsg:
            self.f.write(m)

    def ascape(s): # attribute-escape
        t = str(s)
        for k,v in {'"' : '&quot;', "'" : '&#39;'}:
            t = t.replace(k,v)
        return t
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    
    """ Markup Elements """
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ 
    # NOTE: use self.write(*msg) to put chars to file directly 

    class DUAL_RDR:
        """ implements a fake-handle for dual output - mainly implements write method """
        def __init__(self, parent) -> None:
            self.parent = parent
        def write(self, *args):
            self.parent.f.write(*args)
            self.parent.xf.write(*args) #<--- temporary 'xf' which is sys.stdout
    
    # std-output-redirect

    def rdr_(self, start_as='', dual=False): 
        """ redirects std-out(console output) to log file 
        Args:
            as_code: if True, opens a code block before and after redirecting
            dual:    if True, prints the std-output on consile as well
        """
        self.rdr_as = start_as
        self.tag_(self.rdr_as) if self.rdr_as else None
        self.xf = sys.stdout
        sys.stdout = (LOG.DUAL_RDR(self) if dual else self.f)
    def _rdr(self):
        """ stop redirecting from std-output """
        sys.stdout = self.xf
        self._tag(self.rdr_as) if self.rdr_as else None
        del self.rdr_as
            
    # NewLines
    def nl(self):
        """ [new-line] put newline to file """
        self.f.write('\n')
    def nl2(self):
        """ [new-line-2] put 2 newlines to file """
        self.f.write('\n\n')


    def tag_(self, tag, nl=False, **atr):
        if atr:
            a= ' ' + ' '.join([ '{}="{}"'.format(_name, self.ascape(_value))  for _name, _value in atr.items() ])
        else:
            a= ''
        self.f.write('<' + tag + a + '>')
        self.f.write('\n') if nl else None

    def _tag(self, tag, nl=False):
        self.f.write('</' + tag +'>')
        self.f.write('\n') if nl else None


    # Strings/Chars (Lines and Paragraphs)
    def ln_(self, *msg, sep=' '):
        """ [line] put all items in input to file - seperated by sep, dont end the block """
        for m in msg:
            self.write(m)
            self.f.write(sep)
    def _ln(self, *msg, sep=' '):
        """ [line] put all items in input to file - seperated by sep, dont end the block """
        for m in msg[0:-1]:
            self.write(m)
            self.f.write(sep)
        self.write(msg[-1])

    
    # Lists
    TAG_OLS, TAG_ULS, TAG_LIS = 'ol', 'ul', 'li'
    #--------------------------------------------------------------------------
    def ll(self, L, order=False):
        """ writes items of list L - no nesting """
        tag = self.TAG_OLS if order else self.TAG_ULS
        self.tag_(tag, nl=True)
        for l in L: # for i,l in enumerate(L):
            self.tag_(self.TAG_LIS)
            self.write(l)
            self._tag(self.TAG_LIS, nl=True)
        self._tag(tag, nl=True)
    def ll2(self, H, L, outer_order=False, inner_order=True):
        """ writes items of 2 List with H containing parent and L containing sublist """
        outer_tag = self.TAG_OLS if outer_order else self.TAG_ULS
        #inner_tag = self.OLS if inner_order else self.ULS
        self.tag_(outer_tag, nl=True)
        for h,l in zip(H,L):
            self.tag_(self.TAG_LIS)
            self.write(h)
            self._tag(self.TAG_LIS)
            self.ll(l, order=inner_order)
        self._tag(outer_tag, nl=True)
        
    def lld(self, D, outer_order=False, inner_order=True):
        """ writes items of 2 List with D.keys containing parent and D.values containing sublist """
        self.ll2(list(D.keys()),list(D.values()), outer_order, inner_order)

    
    TAG_TABLE, TAG_TR, TAG_TH, TAG_TD = 'table', 'tr', 'th', 'td'
    # Tables (from iterables)
    def _dump_header(self, header):
        """ helper method, dont use directly """
        self.tag_(self.TAG_TR, nl=True)
        for h in header:
            self.tag_(self.TAG_TH)
            self.write(h)
            self._tag(self.TAG_TH)
        self._tag(self.TAG_TR, nl=True)
    #--------------------------------------------------------------------------
    def t_(self, header):
        """ opens a table """
        self.tag_(self.TAG_TABLE, nl=True)
        self._dump_header(header)

    def r(self, R):
        """ write full row at once """
        self.r_()
        for i in R:
            self.ri(i)
        self._r()
    def r_(self):
        """ open row - for writing unit by unit """
        self.tag_(self.TAG_TR, nl=True)
    def ri(self, i):
        """ write a unit (row item) """
        self.tag_(self.TAG_TD)
        self.write(i)
        self._tag(self.TAG_TD)
    def _r(self):
        """ close a row """
        self._tag(self.TAG_TR, nl=True)
    def _t(self):
        """ close a table """
        self._tag(self.TAG_TABLE, nl=True)
    #--------------------------------------------------------------------------
    def rt(self, header, R):
        """ [Row table] - table with header and each item in R defining one full Row """
        self.t_(header)
        for i in R:
            self.r( i )
        self._t()
    def mrt(self, header, *R):
        """ [Multi-Row table] - table with header and each item in MR defining one full Row
            - auto generates header if its none """
        header = range(len(R[0])) if header is None else header
        self.rt(header, R)
    def ct(self, header, C):
        """ [Col table] - table with header and each item in C defining one full Col """ 
        self.t_(header)
        rows, cols = len(C[0]), len(C)
        for i in range(rows):
            self.r_()
            for j in range(cols):
                self.ri(C[j][i])
            self._r()
        self._t()
    def mct(self, header, *C):
        """ [Multi-Col table] - table with header and each item in C defining one full Col 
        - auto generates header if its none """
        header = range(len(C)) if (header is None) else header
        self.t_(header)
        for cc in zip(*C):
            self.r(cc)
        self._t()
    def dct(self, D):
        """ [Dict col table] - table with header as D.keys() and each item in D.values() defining one full Col 
        - directly calls self.ct with two args - keys and value - as each column """
        self.ct(list(D.keys()), list(D.values()))
    def drt(self, D, hkey='Key', hval='Val'):
        """ [Dict row table] - table with 2-cols (hkey and hval) from a dict 
        - directly calls self.rt with two args - keys and value - as each column """
        self.rt([hkey, hval], D.items() )

    # preformated-text (from dict)
    def pfd(self, D, caption=""):
        """ [pre-formated Dict] - pre-format text from a dict """
        self.c_()
        self.f.write("=-=-=-=-==-=-=-=-=\n"+caption+"\n=-=-=-=-==-=-=-=-=\n")
        for k,v in D.items():
            self.f.write(str(k) + " : " + str(v) + '\n')
        self.f.write("=-=-=-=-==-=-=-=-=\n")
        self._c()

class ESCAPER:
    """ context manager for toggling character escaping while logging,
        > escaping requires extra computing and is avoided by default 
        > user can switch on escaping specific charecters using this context manager """
    def __init__(self, *logs_tup) -> None:
        """ *log_tup : tuples like (log, esc) """
        self.LOT = logs_tup #(log, esc)
    def __enter__(self):
        for log, esc in self.LOT:
            log.pesc = log.esc
            log.escmode(esc)
        return self.LOT[0][0] if len(self.LOT)==1 else tuple( [self.LOT[i][0] for i in self.LOT ])
    def __exit__(self, exc_type, exc_val, exc_tb):
        for log, _ in self.LOT:
            log.escmode(log.pesc)
            del log.pesc
        del self.LOT
        return True

    NBSP_ESC={' ' : '&nbsp;'}
    TAG_ESC = {
    '<' : '&lt;', 
    '>' : '&gt;', 
    }
    STD_ESC = {
    '&' : '&amp;',
    '<' : '&lt;', 
    '>' : '&gt;', 
    }
    ALL_ESC = {
    '&' : '&amp;',
    '<' : '&lt;', 
    '>' : '&gt;', 
    ' ' : '&nbsp;'
    }

""" NOTE:
    * Author:           Nelson.S

"""
#-----------------------------------------------------------------------------------------------------
